"""File containing constants used by CLI."""

# Configuration Constants
CONFIG_PATH_KEY = "MF_CONFIG_DIR"
CONFIG_DWH_DIALECT = "dwh_dialect"
CONFIG_DWH_HOST = "dwh_host"
CONFIG_DWH_PORT = "dwh_port"
CONFIG_DWH_USER = "dwh_user"
CONFIG_DWH_PASSWORD = "dwh_password"
CONFIG_DWH_DB = "dwh_database"
CONFIG_DWH_SCHEMA = "dwh_schema"
CONFIG_DWH_WAREHOUSE = "dwh_warehouse"
CONFIG_MODEL_PATH = "model_path"

# Display Setting Constants
MAX_LIST_OBJECT_ELEMENTS = 5
DEFAULT_RESULT_DECIMAL_PLACES = 2
